"""
Generate JSON Schema for Marshmallow schemas.

"""
from typing import (
    Any,
    Callable,
    Iterable,
    Mapping,
    Set,
    Tuple,
    Type,
)

from marshmallow import Schema
from marshmallow.fields import Field, List, Nested

from microcosm_flask.decorators.schemas import associated_schemas_attr_name
from microcosm_flask.naming import name_for
from microcosm_flask.swagger.naming import type_name


def definition_name_for_schema(schema_cls):
    return type_name(name_for(schema_cls))


class Schemas:
    """
    Swagger schema builder.

    """
    def __init__(
        self,
        build_parameter: Callable[..., Mapping[str, Any]],
        strict_enums: bool = True,
    ):
        self.build_parameter = build_parameter
        self.strict_enums = strict_enums

        # NB: This will break if this class is ever instantiated and then has
        # `build` called more than once
        self.seen_schemas: Set[Type[Schema]] = set()

    def build(self, schema: Schema) -> Mapping[str, Any]:
        """
        Build JSON schema from a marshmallow schema.

        """
        fields = list(self.iter_fields(schema))

        properties = {
            name: self.build_parameter(field, strict_enums=self.strict_enums)
            for name, field in fields
        }

        result = dict(
            type="object",
            properties=properties,
        )

        required_fields = [
            name
            for name, field in fields
            if field.required and not field.allow_none
        ]

        if required_fields:
            result["required"] = required_fields

        return result

    def iter_fields(self, schema: Schema) -> Iterable[Tuple[str, Field]]:
        """
        Iterate through marshmallow schema fields.

        Generates: name, field pairs

        """
        for name in sorted(schema.fields.keys()):
            field = schema.fields[name]
            field_name = field.data_key or name

            yield field_name, field

    def iter_schemas(self, schema: Schema) -> Iterable[Tuple[str, Any]]:
        """
        Build zero or more JSON schemas for a marshmallow schema.

        Generates: name, schema pairs.

        """
        if not schema:
            return

        if type(schema) in self.seen_schemas:
            return

        self.seen_schemas.add(type(schema))

        yield self.to_tuple(schema)

        for associated_schema in getattr(schema, associated_schemas_attr_name(schema.__class__), {}).values():
            yield self.to_tuple(associated_schema())

        for name, field in self.iter_fields(schema):
            if isinstance(field, Nested):
                yield from self.iter_schemas(field.schema)

            if isinstance(field, List) and isinstance(field.inner, Nested):
                yield from self.iter_schemas(field.inner.schema)

    def to_tuple(self, schema: Schema) -> Tuple[str, Any]:
        return definition_name_for_schema(schema), self.build(schema)



"""
Metrics extensions for routes.

"""
from microcosm.api import defaults, typed
from microcosm.config.types import boolean
from microcosm.errors import NotBoundError


@defaults(
    enabled=typed(boolean, default_value=True),
)
class RouteMetrics:

    def __init__(self, graph):
        self.metrics = self.get_metrics(graph)
        self.enabled = bool(
            self.metrics
            and self.metrics.host != "localhost"
            and graph.config.route_metrics.enabled
        )
        self.graph = graph

    def get_metrics(self, graph):
        """
        Fetch the metrics client from the graph.

        Metrics will be disabled if the not configured.

        """
        try:
            return graph.metrics
        except NotBoundError:
            return None

    def __call__(self, endpoint):
        from microcosm_flask.metrics_classifier import StatusCodeClassifier

        def decorator(func):
            key = "route"
            tags = [
                f"endpoint:{endpoint}",
                "backend_type:microcosm_flask",
            ]
            counting = self.graph.metrics_counting(
                key,
                tags=tags,
                classifier_cls=StatusCodeClassifier,
            )
            timing = self.graph.metrics_timing(
                key,
                tags=tags,
            )
            return timing(counting(func))

        return decorator



from dataclasses import dataclass
from os import environ
from typing import Any, Optional
from unittest.mock import MagicMock

from microcosm.decorators import defaults
from microcosm.object_graph import ObjectGraph
from microcosm_logging.decorators import logger


try:
    import sentry_sdk
    from sentry_sdk.utils import BadDsn
except ImportError:
    sentry_sdk = None


@dataclass
class SentryConfig:
    dsn: Optional[str] = None
    enabled: bool = False
    client: Optional[Any] = None


def before_send(event, hint):
    """
    Before sending a event to sentry scrub all values that are non-uuid data from the event
    """
    for exception in event["exception"]["values"]:
        for frame in exception["stacktrace"]["frames"]:
            for var_name, value in frame["vars"].items():
                if var_name in ["self", "cls"]:
                    continue
                elif isinstance(value, str):
                    if value.startswith("<function "):
                        continue
                    elif var_name.endswith("_id"):
                        continue
                frame["vars"][var_name] = "<redacted>"
    return event


@defaults(
    enabled=False,
    dsn=None,  # Data Source Name - should be a url
)
@logger
def configure_sentry(graph: ObjectGraph):
    enabled, dsn, sentry = False, None, None
    if graph.config.sentry_logging.enabled and graph.config.sentry_logging.dsn and sentry_sdk:
        enabled = graph.config.sentry_logging.enabled
        dsn = graph.config.sentry_logging.dsn
        if graph.metadata.testing:
            sentry = MagicMock()
        else:
            try:
                sentry = sentry_sdk.init(
                    graph.config.sentry_logging.dsn,
                    server_name=f"{graph.metadata.name}_api",
                    release=graph.config.build_info_convention.sha1,
                    environment=environ.get("MICROCOSM_ENVIRONMENT") or "undefined",
                    before_send=before_send,
                )
            except BadDsn:
                enabled, dsn, sentry = False, None, None
                configure_sentry.logger.error("Bad DSN value set")
    return SentryConfig(
        dsn=dsn,
        enabled=enabled,
        client=sentry
    )



from microcosm_flask.audit import parse_response
from microcosm_flask.errors import extract_status_code


try:
    from microcosm_metrics.classifier import Classifier
except ImportError:
    raise Exception("Route metrics require 'microcosm-metrics'")


class StatusCodeClassifier(Classifier):
    """
    Label route result/error with its status code.

    """
    def label_result(self, result):
        _, status_code, _ = parse_response(result)
        return self.normalize_status_code(status_code)

    def label_error(self, error):
        status_code = extract_status_code(error)
        return self.normalize_status_code(status_code)

    # Return only the first digit of the status code followed by xx as a string
    # Example: 403 becomes "4xx"
    # This will reduce the cardinality of metrics collected which will reduce cost
    # For detailed status code errors we have logs
    def normalize_status_code(self, status_code: int) -> str:
        return str(status_code)[0] + "xx"



from typing import (
    Any,
    Callable,
    Mapping,
    Optional,
    Sequence,
)

from marshmallow import Schema
from marshmallow.fields import Field

from microcosm_flask.fields import EnumField
from microcosm_flask.swagger.parameters.base import ParameterBuilder


def is_int(value):
    try:
        int(value)
    except Exception:
        return False
    else:
        return True


class EnumParameterBuilder(ParameterBuilder):
    """
    Build an enum parameter.

    If `strict_enums` is True, the parameter is built using the "enum" format,
    with the allowed values under the "enum" key. Otherwise the parameter is simply considered
    as a string.

    """
    def __init__(
        self,
        build_parameter: Callable[[Schema], Mapping[str, Any]],
        strict_enums: bool = True,
        **kwargs,
    ):
        super().__init__(build_parameter)
        self.strict_enums = strict_enums

    def supports_field(self, field: Field) -> bool:
        return bool(getattr(field, "enum", None))

    def parse_format(self, field: Field) -> Optional[str]:
        if isinstance(field, EnumField) and self.strict_enums:
            return "enum"
        return None

    def parse_type(self, field: Field) -> str:
        enum_values = self._parse_enum_values(field)

        if all((isinstance(enum_value, str) for enum_value in enum_values)):
            return "string"
        elif all((is_int(enum_value) for enum_value in enum_values)):
            return "integer"
        else:
            raise Exception(f"Cannot infer enum type for field: {field.name}")

    def _parse_enum_values(self, field: Field) -> Sequence:
        enum = getattr(field, "enum", None)
        return [
            choice.value if field.by_value else choice.name  # type: ignore
            for choice in enum
        ]

    def parse_enum_values(self, field: Field) -> Optional[Sequence]:
        if self.strict_enums:
            return self._parse_enum_values(field)
        return None



"""
Support for encoding and decoding request/response content.

"""
from flask import request
from inflection import camelize
from marshmallow.exceptions import ValidationError
from werkzeug.exceptions import NotFound, UnprocessableEntity

from microcosm_flask.enums import ResponseFormats
from microcosm_flask.naming import name_for


def with_headers(error, headers):
    setattr(error, "headers", headers)
    return error


def with_context(error, errors):
    setattr(error, "context", dict(errors=errors))
    return error


def encode_count_header(count):
    """
    Generate a header for a count HEAD response.

    """
    return {
        "X-Total-Count": count,
    }


def encode_id_header(resource):
    """
    Generate a header for a newly created resource.

    Assume `id` attribute convention.

    """
    if not hasattr(resource, "id"):
        return {}

    return {
        "X-{}-Id".format(
            camelize(name_for(resource))
        ): str(resource.id),
    }


def encode_headers(resource):
    """
    Generate headers from a resource.

    """
    return {}


def load_request_data(request_schema):
    """
    Load request data as JSON using the given schema.

    Forces JSON decoding even if the client not specify the `Content-Type` header properly.

    This is friendlier to client and test software, even at the cost of not distinguishing
    HTTP 400 and 415 errors.

    """
    try:
        json_data = request.get_json(force=True) or {}
    except Exception:
        # if `simplejson` is installed, simplejson.scanner.JSONDecodeError will be raised
        # on malformed JSON, where as built-in `json` returns None
        json_data = {}
    try:
        return request_schema.load(json_data)
    except ValidationError as error:
        raise with_context(
            UnprocessableEntity("Validation error"),
            [
                {
                    "message": "Could not validate field: {}".format(field),
                    "field": field,
                    "reasons": reasons,
                } for field, reasons in error.messages.items()
            ],
        )


def load_query_string_data(request_schema, query_string_data=None):
    """
    Load query string data using the given schema.

    Schemas are assumed to be compatible with the `PageSchema`.

    """
    if query_string_data is None:
        query_string_data = request.args

    try:
        return request_schema.load(query_string_data)
    except ValidationError as error:
        raise with_context(
            UnprocessableEntity("Validation error"),
            error.messages,
        )


def remove_null_values(data):
    if isinstance(data, dict):
        return {
            key: remove_null_values(value)
            for key, value in data.items()
            if value is not None
        }
    if type(data) in (list, tuple):
        return type(data)(map(remove_null_values, data))
    return data


def dump_response_data(response_schema,
                       response_data,
                       status_code=200,
                       headers=None,
                       response_format=None):
    """
    Dumps response data as JSON using the given schema.

    Forces JSON encoding even if the client did not specify the `Accept` header properly.

    This is friendlier to client and test software, even at the cost of not distinguishing
    HTTP 400 and 406 errors.

    """
    if response_schema:
        response_data = response_schema.dump(response_data)

    return make_response(response_data, response_schema, response_format, status_code, headers)


def make_response(response_data,
                  response_schema=None,
                  response_format=None,
                  status_code=200,
                  headers=None,
                  ):

    if response_format is None:
        response_format = ResponseFormats.JSON

    formatter = response_format.value.formatter(response_schema)

    if request.headers.get("X-Response-Skip-Null"):
        # swagger does not currently support null values; remove these conditionally
        response_data = remove_null_values(response_data)

    response = formatter(response_data, headers)
    response.status_code = status_code
    return response


def merge_data(path_data, request_data):
    """
    Merge data from the URI path and the request.

    Path data wins.

    """
    merged = request_data.copy() if request_data else {}
    merged.update(path_data or {})
    return merged


def require_response_data(response_data):
    """
    Enforce that response data is truthy.

    Used to automating 404 errors for CRUD functions that return falsey. Does not
    preclude CRUD functions from raising their own errors.

    :raises NotFound: otherwise

    """
    if not response_data:
        raise NotFound
    return response_data


def find_response_format(allowed_response_formats):
    """
    Basic content negotiation logic.

    If the 'Accept' header doesn't exactly match a format we can handle, we return JSON

    """
    # allowed formats default to [] before this
    if not allowed_response_formats:
        allowed_response_formats = [ResponseFormats.JSON]

    content_type = request.headers.get("Accept")
    if content_type is None:
        # Nothing specified, default to endpoint definition
        if allowed_response_formats:
            return allowed_response_formats[0]
        # Finally, default to JSON
        return ResponseFormats.JSON

    for response_format in ResponseFormats.prioritized():
        if response_format not in allowed_response_formats:
            continue
        if response_format.matches(content_type):
            return response_format

    # fallback for previous behavior
    return ResponseFormats.JSON



"""
Landing Page convention.

"""
from distutils import dist
from io import StringIO
from json import dumps
from pkg_resources import DistributionNotFound, get_distribution

from jinja2 import Template

from microcosm_flask.conventions.registry import iter_endpoints
from microcosm_flask.templates.landing import template


def configure_landing(graph):   # noqa: C901

    def get_properties_and_version():
        """
        Parse the properties from the package information

        """
        try:
            distribution = get_distribution(graph.metadata.name)
            metadata_str = distribution.get_metadata(distribution.PKG_INFO)
            package_info = dist.DistributionMetadata()
            package_info.read_pkg_file(StringIO(metadata_str))
            return package_info
        except DistributionNotFound:
            return None

    def get_swagger_versions():
        """
        Finds all swagger conventions that are bound to the graph

        """
        versions = []

        def matches(operation, ns, rule):
            """
            Defines a condition to determine which endpoints are swagger type

            """
            if(ns.subject == graph.config.swagger_convention.name):
                return True
            return False

        for operation, ns, rule, func in iter_endpoints(graph, matches):
            versions.append(ns.version)

        return versions

    def pretty_dict(dict_):
        return dumps(dict_, sort_keys=True, indent=2, separators=(',', ': '))

    def get_env_file_commands(config, conf_key, conf_string=None):
        if conf_string is None:
            conf_string = []
        for key, value in config.items():
            if isinstance(value, dict):
                get_env_file_commands(value, "{}__{}".format(conf_key, key), conf_string)
            else:
                conf_string.append("export {}__{}='{}'".format(conf_key.upper(), key.upper(), value))
        return conf_string

    def get_links(swagger_versions, properties):
        # add links set in config
        links = {key: value for (key, value) in graph.config.landing_convention.get("links", {}).items()}

        # add links for each swagger version
        for swagger_version in swagger_versions:
            links["swagger {}". format(swagger_version)] = "api/{}/swagger".format(swagger_version)

        # add link to home page
        if hasattr(properties, "url"):
            links["home page"] = properties.url

        return links

    @graph.flask.route("/")
    def render_landing_page():
        """
        Render landing page

        """
        config = graph.config_convention.to_dict()
        env = get_env_file_commands(config, graph.metadata.name)
        health = graph.health_convention.to_dict()
        properties = get_properties_and_version()
        swagger_versions = get_swagger_versions()

        return Template(template).render(
            config=pretty_dict(config),
            description=properties.description if properties else None,
            env=env,
            health=pretty_dict(health),
            links=get_links(swagger_versions, properties),
            service_name=graph.metadata.name,
            version=getattr(properties, 'version', None),
        )



"""
Swagger (OpenAPI) convention.

Exposes swagger definitions for matching operations.

"""
from flask import g
from marshmallow import Schema, fields
from microcosm.api import defaults

from microcosm_flask.conventions.base import Convention
from microcosm_flask.conventions.encoding import load_query_string_data, make_response
from microcosm_flask.conventions.registry import iter_endpoints, request
from microcosm_flask.namespaces import Namespace
from microcosm_flask.operations import Operation
from microcosm_flask.swagger.definitions import build_swagger


class ValidateSwaggerSchema(Schema):
    validate_schema = fields.Boolean()


class SwaggerConvention(Convention):

    @property
    def matching_operations(self):
        return {
            Operation.from_name(operation_name)
            for operation_name in self.graph.config.swagger_convention.operations
        }

    def find_matching_endpoints(self, swagger_ns):
        """
        Compute current matching endpoints.

        Evaluated as a property to defer evaluation.

        """
        def match_func(operation, ns, rule):
            # only expose endpoints that have the correct path prefix and operation
            return (
                rule.rule.startswith(self.graph.build_route_path(swagger_ns.path, swagger_ns.prefix)) and
                operation in self.matching_operations
            )

        return list(iter_endpoints(self.graph, match_func))

    def configure_discover(self, ns, definition):
        """
        Register a swagger endpoint for a set of operations.

        """
        @self.add_route(ns.singleton_path, Operation.Discover, ns)
        @request(ValidateSwaggerSchema)
        def discover():
            request_data = load_query_string_data(ValidateSwaggerSchema())

            swagger = build_swagger(self.graph, ns, self.find_matching_endpoints(ns), **request_data)
            g.hide_body = True
            return make_response(swagger)


@defaults(
    name="swagger",
    operations=[
        "create",
        "create_collection",
        "create_for",
        "delete",
        "replace",
        "replace_for",
        "retrieve",
        "retrieve_for",
        "search",
        "search_for",
        "update",
        "update_batch",
        "update_for",
        "upload",
        "upload_for",
    ],
    version="",
)
def configure_swagger(graph):
    """
    Build a singleton endpoint that provides swagger definitions for all operations.

    """
    ns = Namespace(
        subject=graph.config.swagger_convention.name,
        version=graph.config.swagger_convention.version,
    )
    convention = SwaggerConvention(graph)
    convention.configure(ns, discover=tuple())
    return ns.subject



from microcosm_flask.swagger.parameters.decorated import SWAGGER_FORMAT, SWAGGER_TYPE


def swagger_field(swagger_type="string", swagger_format=None):
    """
    Decorator for an existing field type to coerce a specific swagger type/format.

    Usage:

        class MySchema(Schema):
             myfield = swagger_field(swagger_type="string")(fields.Foo())

    """
    def decorator(field):
        setattr(field, SWAGGER_TYPE, swagger_type)
        setattr(field, SWAGGER_FORMAT, swagger_format)
        return field

    return decorator



"""
API interfaces for swagger operations.

"""
from typing import (
    Any,
    Iterable,
    Mapping,
    Tuple,
)

from marshmallow import Schema
from marshmallow.fields import Field

from microcosm_flask.swagger.parameters import Parameters
from microcosm_flask.swagger.schemas import Schemas


def build_schema(schema: Schema, strict_enums: bool = True) -> Mapping[str, Any]:
    """
    Build JSON schema from a marshmallow schema.

    """
    builder = Schemas(build_parameter=build_parameter, strict_enums=strict_enums)
    return builder.build(schema)


def iter_schemas(schema: Schema, strict_enums: bool = True) -> Iterable[Tuple[str, Any]]:
    """
    Build zero or more JSON schemas for a marshmallow schema.

    Generates: name, schema pairs.

    """
    builder = Schemas(build_parameter=build_parameter, strict_enums=strict_enums)
    return builder.iter_schemas(schema)


def build_parameter(field: Field, **kwargs) -> Mapping[str, Any]:
    """
    Build JSON parameter from a marshmallow field.

    """
    builder = Parameters(**kwargs)
    return builder.build(field)



"""
A discovery endpoint provides links to other endpoints.

"""
from microcosm.api import defaults

from microcosm_flask.conventions.base import Convention
from microcosm_flask.conventions.encoding import make_response
from microcosm_flask.conventions.registry import iter_endpoints
from microcosm_flask.linking import Link, Links
from microcosm_flask.namespaces import Namespace
from microcosm_flask.operations import Operation
from microcosm_flask.paging import OffsetLimitPage, OffsetLimitPageSchema


def iter_links(operations, page):
    """
    Generate links for an iterable of operations on a starting page.

    """
    for operation, ns, rule, func in operations:
        yield Link.for_(
            operation=operation,
            ns=ns,
            type=ns.subject_name,
            qs=page.to_items(),
        )


class DiscoveryConvention(Convention):

    @property
    def matching_operations(self):
        return {
            Operation.from_name(operation_name)
            for operation_name in self.graph.config.discovery_convention.operations
        }

    def find_matching_endpoints(self, discovery_ns):
        """
        Compute current matching endpoints.

        Evaluated as a property to defer evaluation.

        """
        def match_func(operation, ns, rule):
            return operation in self.matching_operations

        return list(iter_endpoints(self.graph, match_func))

    def configure_discover(self, ns, definition):
        """
        Register a discovery endpoint for a set of operations.

        """
        page_schema = OffsetLimitPageSchema()

        @self.add_route("/", Operation.Discover, ns)
        def discover():
            # accept pagination limit from request
            page = OffsetLimitPage.from_query_string(page_schema)
            page.offset = 0

            response_data = dict(
                _links=Links({
                    "self": Link.for_(Operation.Discover, ns, qs=page.to_items()),
                    "search": [
                        link for link in iter_links(self.find_matching_endpoints(ns), page)
                    ],
                }).to_dict()
            )
            return make_response(response_data)


@defaults(
    name="hal",
    operations=[
        "search",
    ],
)
def configure_discovery(graph):
    """
    Build a singleton endpoint that provides a link to all search endpoints.

    """
    ns = Namespace(
        subject=graph.config.discovery_convention.name,
    )
    convention = DiscoveryConvention(graph)
    convention.configure(ns, discover=tuple())
    return ns.subject



from datetime import datetime, timedelta
from logging import Logger
from tracemalloc import start, take_snapshot

from microcosm.api import binding, defaults, typed
from microcosm.config.types import boolean


@binding("memory_profiler")
@defaults(
    enabled=typed(boolean, False),
    report_size_lines=typed(int, 10),
    sampling_interval_min=typed(int, 10),
)
class MemoryProfiler:
    logger: Logger

    def __init__(self, graph):
        self.enabled = graph.config.memory_profiler.enabled
        self.report_size_lines = graph.config.memory_profiler.report_size_lines
        self.logger = graph.logger

        self.sampling_interval_min = graph.config.memory_profiler.sampling_interval_min
        self.last_sampling_time_delta = timedelta(minutes=self.sampling_interval_min)

        self.last_sampling_time = datetime.now()

        if not self.enabled:
            self.logger.info("Skipping initialization because memory profiling is not enabled!")
            return

        start()
        self.origin_snapshot = take_snapshot()

    def snapshot_at_intervals(self, func):
        def maybe_snapshot(*args, **kwargs):
            result = func(*args, **kwargs)
            if not self.enabled:
                self.logger.debug("Memory profiling is disabled. Please enable to get snapshots.")
                return result

            now = self.get_now()
            if now - self.last_sampling_time > self.last_sampling_time_delta:
                self.take_snapshot(now)

            return result
        return maybe_snapshot

    def take_snapshot(self, current_time):
        latest_snapshot = take_snapshot()
        difference = latest_snapshot.compare_to(self.origin_snapshot, "lineno")
        top_ten = difference[:self.report_size_lines]
        worst_offenders = "\n".join([str(memory_item) for memory_item in top_ten])
        self.logger.info(f"Memory snapshot: \n {worst_offenders}")
        self.last_sampling_time = current_time

    def get_now(self):
        """
        Wrap datetime.now for easier mocking.
        """
        return datetime.now()
